import bpy
import bmesh
from ... import __package__ as base_package

def edit_mesh_node_group():
    edit_mesh = bpy.data.node_groups.new(type = 'GeometryNodeTree', name = "Edit Mesh")

    edit_mesh.color_tag = 'NONE'
    edit_mesh.description = ""
    edit_mesh.default_group_node_width = 140
    edit_mesh.is_modifier = True

    #edit_mesh interface
    #Socket Geometry
    geometry_socket = edit_mesh.interface.new_socket(name = "Geometry", in_out='OUTPUT', socket_type = 'NodeSocketGeometry')
    geometry_socket.attribute_domain = 'POINT'

    #Socket Geometry
    geometry_socket_1 = edit_mesh.interface.new_socket(name = "Geometry", in_out='INPUT', socket_type = 'NodeSocketGeometry')
    geometry_socket_1.attribute_domain = 'POINT'

    #initialize edit_mesh nodes
    #node Group Input
    group_input = edit_mesh.nodes.new("NodeGroupInput")
    group_input.name = "Group Input"

    #Socket Index
    index_socket = edit_mesh.interface.new_socket(name = "Index", in_out='INPUT', socket_type = 'NodeSocketInt')
    index_socket.default_value = 0
    index_socket.min_value = -2147483648
    index_socket.max_value = 2147483647
    index_socket.subtype = 'NONE'
    index_socket.attribute_domain = 'POINT'
    
    #node Group Output
    group_output = edit_mesh.nodes.new("NodeGroupOutput")
    group_output.name = "Group Output"
    group_output.is_active_output = True

    #Set locations
    group_input.location = (-340.0, 0.0)
    group_output.location = (200.0, 0.0)

    #initialize edit_mesh links
    #group_input.Geometry -> group_output.Geometry
    edit_mesh.links.new(group_input.outputs[0], group_output.inputs[0])
    return edit_mesh

def apply_modifiers(obj):
    if obj:
        if obj.type == 'MESH':
            depsgraph = bpy.context.evaluated_depsgraph_get()
            object_eval = obj.evaluated_get(depsgraph)
            mesh_from_eval = bpy.data.meshes.new_from_object(object_eval, depsgraph=depsgraph)

            if obj.mode == 'OBJECT':
                obj.data = mesh_from_eval
            else:
                bm = bmesh.from_edit_mesh(obj.data)
                bm.clear()
                bm.from_mesh(mesh_from_eval)
                bmesh.update_edit_mesh(obj.data)

            for modifier in obj.modifiers: # hide all modifiers
                modifier.show_viewport = False

            object_eval.to_mesh_clear() #clear the evaluated mesh
            return obj

def draw_edit_mesh_modifier(self, context):
    prefs = bpy.context.preferences.addons[base_package].preferences
    obj = context.active_object
 
    if prefs.properties_editor_style == 'LIST' and obj and obj.type == 'MESH': # only works in list mode
        layout = self.layout
        layout.operator("object.edit_mesh_modifier", icon='EDITMODE_HLT')

class EditmeshClear(bpy.types.Operator):
    bl_idname = "object.edit_mesh_clear"
    bl_label = "Clear edit mesh"
    bl_description = "Clear edit mesh Modifier"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        active_mod_index = obj.ml_modifier_active_index
        active_mod = obj.modifiers[active_mod_index]
        property_index = str(active_mod["Socket_2"])

        if obj and active_mod:
            if active_mod.type == 'NODES':
                if active_mod.node_group.name == "Edit Mesh":
                    amount_of_edit_mesh_modifiers = 0
                    old_mesh_data = obj.data

                    obj.modifiers.remove(active_mod) # remove Edit Mesh Modifier

                    object_mode = True
                    if context.mode != 'OBJECT':
                        object_mode = False
                        bpy.ops.object.editmode_toggle()

                    prop_name = "edit_mesh_data_" + str(property_index)  # get the obj.data index 

                    if prop_name in obj:
                        data = (obj[prop_name])
                        obj.data = data # restore the original mesh data

                        del obj[prop_name]
                        
                        if not object_mode:
                            bpy.ops.ed.undo_push()
                            bpy.ops.object.editmode_toggle()
                      
                        # restore modifiers, we only do it until the next Edit Mesh Modifier!
                        for modifier in reversed(obj.modifiers):
                            if modifier.type == 'NODES':
                                if modifier.node_group and modifier.node_group.name == "Edit Mesh":
                                    break # we reached another Edit Mesh Modifier, break loop
                                else:
                                    modifier.show_viewport = True
                            else:
                                modifier.show_viewport = True

                        bpy.data.meshes.remove(old_mesh_data) # delete old mesh data

                    else:
                        self.report({'ERROR'}, "No mesh data was found")
                        return {'CANCELLED'}
                    
                    # cleanup obj properites if no edit mesh modifiers
                    for m in obj.modifiers:
                        if m.type == 'NODES' and m.node_group:
                            if "Edit Mesh" in m.node_group.name:
                                amount_of_edit_mesh_modifiers += 1
                    
                    # if no edit mesh mdifers left, cleanup and delete all edit_mesh_data_ obj properties, just in case
                    if amount_of_edit_mesh_modifiers == 0:
                        for prop_name in obj.keys():
                            if prop_name.startswith("edit_mesh_data_"):
                                # old_data = (obj[prop_name]) # might not be safe, check amount of users first?
                                # bpy.data.meshes.remove(old_data) # delete old mesh data
                                del obj[prop_name]
   
        return {'FINISHED'}


class EditmeshModifier(bpy.types.Operator):
    bl_idname = "object.edit_mesh_modifier"
    bl_label = "Edit Mesh"
    bl_description = "Edit mesh Modifier"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
    
        if obj.mode == 'EDIT':
            obj.update_from_editmode() # update mesh data, so it get latest obj.data in edit mode withoute needing to toggle object mode
           
        backup_mesh_data = obj.data.copy()
        backup_mesh_data.use_fake_user = True
        base_name = obj.data.name.split("_edit_mesh_")[0]
        backup_mesh_data.name = base_name + "_edit_mesh_" + str(len([block for block in bpy.data.meshes if block.name.startswith(base_name + "_edit_mesh_")]))

        index = 0
        while "edit_mesh_data_" + str(index) in obj: # make sure obj property does not exist already
            index += 1
        
        # obj["edit_mesh_data_" + str(index)] = backup_mesh_data # this one is not hidden by deafult, cluttering up the panel. Also worse for debugging
        prop_name = "edit_mesh_data_" +  str(index)
        setattr(bpy.types.Object, prop_name, bpy.props.PointerProperty(
            name="Edit Mesh Data" + str(index), 
            description=str(index),
            type=bpy.types.Mesh
        ))
        
        setattr(obj, prop_name, backup_mesh_data) # Assign the custom property to the object
        obj = apply_modifiers(obj) # get evaulted mesh

        if not "Edit Mesh" in bpy.data.node_groups:
            edit_mesh_node_group()

        edit_mesh_modifier = obj.modifiers.new(name="Edit Mesh", type='NODES')
        edit_mesh_modifier.node_group = bpy.data.node_groups['Edit Mesh']
        edit_mesh_modifier.show_group_selector = False
        edit_mesh_modifier["Socket_2"] = index # we store a index of the number of the mesh data block, to later use when restoring it
        edit_mesh_modifier.show_viewport = False # no need to calculate the geomtry node, its just a placholder modifer that get ovveriden UI wise

        return {'FINISHED'}
        
    def register():
        bpy.types.OBJECT_MT_modifier_add_edit.prepend(draw_edit_mesh_modifier)

    def unregister():
        bpy.types.OBJECT_MT_modifier_add_edit.remove(draw_edit_mesh_modifier)
